import astroid
from pylint import interfaces
from pylint.checkers import utils

from .base_checker import BaseChecker
from .. import messages


class FieldChecker(BaseChecker):
    __implements__ = interfaces.IAstroidChecker

    msgs: dict = messages.FIELD_CHECKER_MSGS
    validators: dict = {
        'Many2one': {
            're': '^.*_id$',
            'msg_id': 'biszx-relation2one-field-name'
        },
        'Many2one_user': {
            're': '^.*_uid|^.*_id$',
            'msg_id': 'biszx-relation2one-field-name'
        },
        'One2many': {
            're': '^.*_ids$',
            'msg_id': 'biszx-relation2many-field-name'
        },
        'One2many_user': {
            're': '^.*_uids|^.*_ids$',
            'msg_id': 'biszx-relation2many-field-name'
        },
        'Many2many': {
            're': '^.*_ids$',
            'msg_id': 'biszx-relation2many-field-name'
        },
        'Many2many_user': {
            're': '^.*_uids|^.*_ids$',
            'msg_id': 'biszx-relation2many-field-name'
        },
        'Boolean': {
            're': '^is_*|^active$',
            'msg_id': 'biszx-boolean-field-name'
        },
        'Date': {
            're': '^.*_date$',
            'msg_id': 'biszx-date-field-name'
        },
        'Datetime': {
            're': '^.*_datetime$',
            'msg_id': 'biszx-datetime-field-name'
        },
    }

    @utils.check_messages('biszx-relation2one-field-name',
                          'biszx-relation2many-field-name',
                          'biszx-boolean-field-name',
                          'biszx-date-field-name',
                          'biszx-datetime-field-name')
    def visit_assign(self, node: astroid.Assign):
        find: tuple = tuple(self.validators.keys())
        value = node.value
        if (
            value
            and isinstance(value, astroid.Call)
            and isinstance(value.func, astroid.Attribute)
            and isinstance(value.func.expr, astroid.Name)
            and value.func.expr.name == 'fields'
            and (key := value.func.attrname) in find
        ):
            if self._is_relation_user(key, value):
                self._validate_match_name(
                    node,
                    f'{key}_user',
                    node.targets[0].name
                )
            else:
                self._validate_match_name(
                    node,
                    key,
                    node.targets[0].name
                )

    def _is_relation_user(self, key: str, value: astroid.Call) -> bool:
        return (
            key in ('Many2one', 'One2many', 'Many2many')
            and any([
                (
                    len(value.args)
                    and getattr(value.args[0], 'value', None) == 'res.users'
                ),
                (
                    len(value.keywords)
                    and (
                        model := list(
                            filter(lambda v: v.arg == 'comodel_name',
                                   value.keywords)
                        )
                    )
                    and getattr(model[0].value, 'value', None) == 'res.users'
                )
            ])
        )



from .. import settings

FUNCTION_CHECKER_MSGS = {
    f'C{settings.FUNCTION_MSG_ID}01': (
        'Default function name must be start with _default_',
        'biszx-default-func-name',
        settings.MSG_DESCRIPTION
    ),
    f'C{settings.FUNCTION_MSG_ID}02': (
        'Search function name must be start with _search_',
        'biszx-search-func-name',
        settings.MSG_DESCRIPTION
    ),
    f'C{settings.FUNCTION_MSG_ID}03': (
        'Compute function name must be start with _compute_',
        'biszx-compute-func-name',
        settings.MSG_DESCRIPTION
    ),
    f'C{settings.FUNCTION_MSG_ID}04': (
        'Onchange function name must be start with _onchange_',
        'biszx-onchange-func-name',
        settings.MSG_DESCRIPTION
    ),
    f'C{settings.FUNCTION_MSG_ID}05': (
        'Constrains function name must be start with _check_',
        'biszx-constrains-func-name',
        settings.MSG_DESCRIPTION
    ),
    f'C{settings.FUNCTION_MSG_ID}06': (
        'Domain function name must be start with _domain_',
        'biszx-domain-func-name',
        settings.MSG_DESCRIPTION
    ),
    f'C{settings.FUNCTION_MSG_ID}07': (
        'Inverse function name must be start with _inverse_',
        'biszx-inverse-func-name',
        settings.MSG_DESCRIPTION
    ),
}



from .checker import FunctionChecker, FieldChecker
from . import messages


def register(linter):
    '''required method to auto register this checker'''
    linter.register_checker(FunctionChecker(linter))
    linter.register_checker(FieldChecker(linter))


def get_all_messages():
    '''Get all messages of this plugin'''
    all_msgs = {}
    all_msgs.update(messages.FUNCTION_CHECKER_MSGS)
    all_msgs.update(messages.FIELD_CHECKER_MSGS)
    return all_msgs


def messages2md():
    all_msgs = get_all_messages()
    md_msgs = 'Code | Description | short name\n--- | --- | ---'
    for msg_code, (title, name_key, _) in sorted(all_msgs.items()):
        title = title.replace('_', '\\_')
        md_msgs += f'\n{msg_code} | {title} | {name_key}'
    md_msgs += '\n'
    return md_msgs



from .. import settings

FIELD_CHECKER_MSGS = {
    f'C{settings.FIELD_MSG_ID}01': (
        'Many2one field name must be end with _id or _uid for res.users',
        'biszx-relation2one-field-name',
        settings.MSG_DESCRIPTION
    ),
    f'C{settings.FIELD_MSG_ID}02': (
        ('One2many and Many2many field name must be end with _ids'
         ' or _uids for res.users'),
        'biszx-relation2many-field-name',
        settings.MSG_DESCRIPTION
    ),
    f'C{settings.FIELD_MSG_ID}03': (
        'Boolean field name must be start with is_',
        'biszx-boolean-field-name',
        settings.MSG_DESCRIPTION
    ),
    f'C{settings.FIELD_MSG_ID}04': (
        'Date field name must be end with _date',
        'biszx-date-field-name',
        settings.MSG_DESCRIPTION
    ),
    f'C{settings.FIELD_MSG_ID}05': (
        'Datetime field name must be end with _at',
        'biszx-datetime-field-name',
        settings.MSG_DESCRIPTION
    ),
}



import re
from typing import Any
from pylint.checkers import BaseChecker

from .. import settings


class BaseChecker(BaseChecker):
    name: str = settings.CHECKER_NAME

    def _validate_match_name(
        self,
        node: Any,
        key: str,
        name: str,
    ) -> None:
        value: dict = self.validators[key]
        if not re.match(value['re'], name):
            self.add_message(value['msg_id'], node=node)
